@@ -18,20 +18,27 @@
 __all__ = [
     'CMakeClient',
     'CMakeExecutor',
+    'CMakeExecScope',
     'CMakeException',
     'CMakeFileAPI',
     'CMakeInterpreter',
     'CMakeTarget',
+    'CMakeToolchain',
     'CMakeTraceLine',
     'CMakeTraceParser',
+    'SingleTargetOptions',
+    'TargetOptions',
     'parse_generator_expressions',
     'language_map',
+    'cmake_defines_to_args',
+    'check_cmake_args',
 ]
 
-from .common import CMakeException
+from .common import CMakeException, SingleTargetOptions, TargetOptions, cmake_defines_to_args, language_map, check_cmake_args
 from .client import CMakeClient
 from .executor import CMakeExecutor
 from .fileapi import CMakeFileAPI
 from .generator import parse_generator_expressions
-from .interpreter import CMakeInterpreter, language_map
+from .interpreter import CMakeInterpreter
+from .toolchain import CMakeToolchain, CMakeExecScope
 from .traceparser import CMakeTarget, CMakeTraceLine, CMakeTraceParser
